// SpectraV2_GeneralCNF.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <iostream>
#include <string>
#include <algorithm>
#include "MonoCNFInstance.h"
#include "SimulationUtils.h"

#include "boost/program_options.hpp" 

using namespace std;


vector<double> GetComullativeBinomials(int n);
vector<double> GetComullativeDistrF(vector<double>& spectra);
double Bin(int n,int k);

double EstimateMonoCover(int N, double p, MonoCNFInstance& instance);
double EstimateMonoCoverUnbiasedEstimator(int N, double p, MonoCNFInstance& instance, SimulationUtils& utils,double& RE);
int PerformSingleRunMC(vector<int>& perm, MonoCNFInstance& graph);

void PrintVector(vector<double>& spectra);

int main(int argc, char* argv[])
{

	string problem_path; 
	int N = 1000;
	int R = 10;
	namespace po = boost::program_options;

	po::options_description desc("Options"); 
    desc.add_options() 
      ("help", "Specify the problem in valid cnf format and the sample size. Example --problem problemname.cnf --N 10000 --R 10") 
      ("problem", po::value<std::string>(&problem_path), "problem (path) to run")
	  ("N",  po::value<int>(&N), "Sample size")
	  ("R",  po::value<int>(&R), "Number of independent experiments"); 


	po::variables_map vm; 
    try 
    { 
		  po::store(po::parse_command_line(argc, argv, desc),  
					vm); // can throw 
 
		  /** --help option 
		   */ 
		  if ( vm.count("help")  ) 
		  { 
			std::cout << "Basic Command Line Parameter App" << std::endl 
					  << desc << std::endl; 
			return 0; 
		  } 

		if (vm.count("problem") == 0)
		{
			cout << "No problem specified" << endl;
			return 1;
		}

		if (vm.count("N") == 0)
		{
			cout << "No sample size specified" << endl;
			return 1;
		}

		if (vm.count("R") == 0)
		{
			cout << "No replication number specified" << endl;
			return 1;
		}
 
		po::notify(vm); // throws on error, so do after help in case 
						  // there are any problems 
	} 
	catch(po::error& e) 
	{ 
		  std::cerr << "ERROR: " << e.what() << std::endl << std::endl; 
		  std::cerr << desc << std::endl; 
		  return -1; 
	} 
	
	try{
		SimulationUtils utils;

		MonoCNFInstance instance;
		instance.Load(problem_path);


		vector<double> res;
		vector<double> RE_ARR;
		vector<double> cputime;
		cout<<endl<<"#replication | estimator | RE | CPU"<<endl<<"-----------------------------------------"<<endl; 
		for (int i=0;i<R; i++)
		{
			double RE = 0;
			utils.StartCPUCount();
			//double l = EstimateMonoCover( N, 0.5, instance);
			double l = EstimateMonoCoverUnbiasedEstimator( N, 0.5, instance, utils, RE);
			RE_ARR.push_back(RE);
			double cpu = utils.GetCPUCount();
			cout<<i+1<<" "<<l<<" "<<RE<<" "<< cpu<<endl;
			cputime.push_back(cpu);
			res.push_back(l);
		}
		cout<<"-----------------------------------------"<<endl;
		cout<<"Average: "<<utils.Mean(res)<<" "<<utils.Mean(RE_ARR)<<" "<<utils.Mean(cputime)<<endl;

		//double anotherRER = utils.GetRelativeError(res)*sqrt((double)R);
		//cout<<"another RE = "<<anotherRER<<endl;
		} 
		catch(std::exception& e) 
		{ 
			cout << "Unhandled Exception reached the top of main: " 
				  << e.what() << ", application will now exit" << std::endl; 
			return 0; 
		} 
	return 0;
}

vector<double> GetComullativeBinomials(int n)
{
	vector<double> bin;
	for(int i=0;i<n+1;i++)
	{
		bin.push_back(Bin(n,i));
	}

	vector<double> res;
	for(int i=0;i<n+1; i++)
	{
		res.push_back(0);
		for(int j=i;j<=n;j++)
		{
			res[i] += bin[j];
		}
	}

	return res;
}

void PrintVector(vector<double>& spectra)
{
	cout<<"---------------------------------"<<endl;
	for(int i=0; i<spectra.size(); i++)
		cout<<" "<<spectra[i];

	cout<<endl;
	cout<<"---------------------------------"<<endl;
}


double EstimateMonoCoverUnbiasedEstimator(int N, double p, MonoCNFInstance& instance,SimulationUtils& utils,double& RE)
{

	int sz = instance.GetVarsNum();
	vector<double> F(sz+1);
	F.assign(sz+1,0);

	vector<double> vec(sz+1);
	vec.assign(sz+1,0);

	vector<double> comBin = GetComullativeBinomials(sz);
	vector<int> vars; 
	for(int i=0;i<sz; i++)
		vars.push_back(i);

	vector<double> X;
	double X_SQUARE = 0;
	for (int i=0;i<N; i++)
	{
		int ind = PerformSingleRunMC(vars,instance);
		F[ind]++;
		X.push_back(comBin[ind+1]);
	}

	RE=utils.GetRelativeError(X);
	double ell = utils.Mean(X);

	return  ell;
}

double EstimateMonoCover(int N, double p, MonoCNFInstance& instance)
{
	int sz = instance.GetVarsNum();
	vector<int> vars; 
	for(int i=0;i<sz; i++)
		vars.push_back(i);

	vector<double> vec(sz+1);
	vec.assign(sz+1,0);


	for (int i=0;i<N; i++)
	{
		int ind = PerformSingleRunMC(vars,instance);
		vec[ind+1]++;
	}

	// normalize
	for(int i=0;i<vec.size(); i++)
	{
		vec[i]/=(double)N;
	}

	PrintVector(vec);
	// calculate cdf
	vec = GetComullativeDistrF(vec);

	//PrintVector(vec);

	// get an estimator
	double l = 0;
	int n=vec.size()-1;
	for (int i=0; i<=n; i++)
	{
		l+=Bin(n,i)*vec[i];
	}

	return l;
}

int PerformSingleRunMC(vector<int>& perm, MonoCNFInstance& instance)
{
	random_shuffle(perm.begin(),perm.end());	
	int b = instance.GetCriticalIndex(perm);
	return b;
}

vector<double> GetComullativeDistrF(vector<double>& spectra)
{
	vector<double> ComF;
	ComF.push_back(spectra[0]);
	for(int i=1;i<spectra.size(); i++)
	{
		double prev = ComF[i-1];
		double newV = prev + spectra[i];
		ComF.push_back(newV);
	}

	return ComF;
}

double Bin(int n,int k)
{
	double res = 1;
	for (int i=1; i<= min(k,n-k); i++)
	{
		res = res*(n-i+1);
		res = res/i;
	}
	return res;
}